/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.onlineSim.resilience;

import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.interfaces.simulation.IEventProcessor;
import com.net2plan.interfaces.simulation.SimAction;
import com.net2plan.interfaces.simulation.SimEvent;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.libraries.TrafficComputationEngine;
import com.net2plan.utils.Pair;
import com.net2plan.utils.StringUtils;
import com.net2plan.utils.Triple;
import java.util.EnumSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

/**
 * <p>Algorithm which implements several restoration mechanisms.</p>
 *
 * <p>For any of the implemented mechanisms, we use the following terminology:</p>
 * <ul>
 * <li>Point of local repair (PLR): Node that redirectes the traffic onto the backup path</li>
 * <li>Merge point (MP): Downstream node where the backup path rejoins to the primary path</li>
 * </ul>
 * 
 * <p>Implemented mechanisms:</p>
 * <ul>
 * <li>Path restoration: Try to find a path with enough capacity from the ingress node (PLR) to the egress node (MP)</li>
 * <li>Subpath restoration: Try to find a path with enough capacity from the first available node before any failure (PLR) to the first available node after any failure (MP)</li>
 * <li>Link restoration: Try to find a path with enough capacity from the first available node before a failure (PLR) to the first available node after a failure (MP). In case of multiple failures, multiple link restorations are tried sequentially</li>
 * <li>Fast reroute: Try to find a path with enough capacity from the first available node before any failure (PLR) to the egress node (MP)</li>
 * </ul>
 * 
 * <p><b>Important</b>: Local restoration mechanisms (all but path restoration) considers 
 * as spare capacity that available after reducing the amount of traffic expected to be 
 * carried through ingressNode-to-PLR and MP-to-egressNode segments, since some loops may appear.</p>
 * 
 * <p>Upon reparation, try to restore each rerouted connection to their original path, 
 * using a make-before-break strategy, only in case 'revertiveMode' is enabled. Then,
 * still-failing routes are tried to be restored as under failure events.</p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class NRSim_AA_genericRestorationAlgorithm extends IEventProcessor
{
	private double PRECISION_FACTOR;
	private boolean revertiveMode;
	private String restorationMode;
	
	@Override
	public String getDescription()
	{
		String NEW_LINE = StringUtils.getLineSeparator();
		StringBuilder description = new StringBuilder();
		description.append("This algorithm implements several restoration mechanisms").append(NEW_LINE).append(NEW_LINE);
		description.append("For any of the implemented mechanisms, we use the following terminology:").append(NEW_LINE);
		description.append("- Point of local repair (PLR): Node that redirectes the traffic onto the backup path").append(NEW_LINE);
		description.append("- Merge point (MP): Downstream node where the backup path rejoins to the primary path").append(NEW_LINE).append(NEW_LINE);
		description.append("Implemented mechanisms:").append(NEW_LINE);
		description.append("- Path restoration: Try to find a path with enough capacity from the ingress node (PLR) to the egress node (MP)").append(NEW_LINE);
		description.append("- Subpath restoration: Try to find a path with enough capacity from the first available node before any failure (PLR) to the first available node after any failure (MP)").append(NEW_LINE);
		description.append("- Link restoration: Try to find a path with enough capacity from the first available node before a failure (PLR) to the first available node after a failure (MP). In case of multiple failures, multiple link restorations are tried sequentially").append(NEW_LINE);
		description.append("- Fast reroute: Try to find a path with enough capacity from the first available node before any failure (PLR) to the egress node (MP)").append(NEW_LINE).append(NEW_LINE);
		description.append("Important: Local restoration mechanisms (all but path restoration) considers as spare capacity that available after reducing the amount of traffic expected to be carried through ingressNode-to-PLR and MP-to-egressNode segments, since some loops may appear.").append(NEW_LINE).append(NEW_LINE);
		description.append("Upon reparation, try to restore each rerouted connection to their original path, using a make-before-break strategy, only in case 'revertiveMode' is enabled. Then, still-failing routes are tried to be restored as under failure events.");
		
		return description.toString();
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> parameters = new LinkedList<Triple<String, String, String>>();
		parameters.add(Triple.of("revertiveMode", "#boolean# true", "Indicates whether routes are restored to their primary paths when become available"));
		parameters.add(Triple.of("restorationMode", "#select# pathRestoration subpathRestoration linkRestoration fastReroute", "Restoration mode: path restoration (from ingress node to egress node), subpathRestoration (from the first node before the first failure to the first node after the last failure), linkRestoration (from the first node before a failure to the next available one), fastReroute (from first node before the failure to egress node)"));
		
		return parameters;
	}

	@Override
	public void initialize(NetPlan initialNetPlan, NetPlan currentNetPlan, Map<String, String> algorithmParameters, Map<String, String> simulationParameters, Map<String, String> net2planParameters)
	{
		int N = currentNetPlan.getNumberOfNodes();
		int E = currentNetPlan.getNumberOfLinks();
		int D = currentNetPlan.getNumberOfDemands();
		int R = currentNetPlan.getNumberOfRoutes();
		if (N == 0 || E == 0 || D == 0 || R == 0) throw new Net2PlanException("A physical topology (nodes and links), demands and routing are required");
		if (currentNetPlan.getLinkMaximumUtilization() > 1) throw new Net2PlanException("Designs with over-subscribed links are not valid");
			
		PRECISION_FACTOR = Double.parseDouble(net2planParameters.get("precisionFactor"));
		revertiveMode = Boolean.parseBoolean(algorithmParameters.get("revertiveMode"));
		restorationMode = algorithmParameters.get("restorationMode");
		switch(restorationMode)
		{
			case "pathRestoration":
			case "subpathRestoration":
			case "linkRestoration":
			case "fastReroute":
				break;
				
			default:
				throw new Net2PlanException("Bad - Invalid restoration mode");
		}
	}

	@Override
	public void processEvent(NetPlan currentNetPlan, SimEvent event)
	{
		/* Check event type */
		boolean isFailureEvent = false;
		boolean isReparationEvent = false;
		if (event.containsAny(EnumSet.of(SimAction.ActionType.NODE_DOWN, SimAction.ActionType.LINK_DOWN))) isFailureEvent = true;
		if (event.containsAny(EnumSet.of(SimAction.ActionType.NODE_UP, SimAction.ActionType.LINK_UP))) isReparationEvent = true;
		if (isFailureEvent && isReparationEvent) throw new Net2PlanException("Failures and reparations cannot be processed as a single event");
		if (!isFailureEvent && !isReparationEvent) throw new Net2PlanException("This algorithm only processes failure and reparation events");
		
		if (isFailureEvent)
		{
			/* Try to reroute as much failing traffic as possible */
			rerouteFailingTraffic(currentNetPlan);
		}
		else
		{
			/* If revertiveMode is applied, then try to restore to the original path only those up routes */
			if (revertiveMode) restoreToPrimary(currentNetPlan);
			
			/* Try to reroute as much failing traffic as possible */
			rerouteFailingTraffic(currentNetPlan);
		}
	}
	
	/**
	 * Computes the link spare capacity map considering that capacity from 
	 * ingressNode-to-LRP and MP-to-egressNode is actually 
	 * reserved.
	 * 
	 * @param currentNetPlan Current network state
	 * @param path Sequence of links of the failing path
	 * @param localRepairPoint Node where initiates the backup path
	 * @param mergePoint Node where terminates the backup path
	 * @param x_p Carried traffic
	 * @return Spare capacity per link (zero for failing links)
	 * @since 1.0
	 */
	private static Map<Long, Double> actualLinkSpareCapacityMap(NetPlan currentNetPlan, List<Long> path, long localRepairPoint, long mergePoint, double x_p)
	{
		Map<Long, Double> linkSpareCapacityMap = currentNetPlan.getLinkSpareCapacityMap();
		
		ListIterator<Long> it = path.listIterator();
		while(it.hasNext())
		{
			long linkId = it.next();
			Pair<Long, Long> nodePair_thisLink = currentNetPlan.getLinkOriginDestinationNodePair(linkId);
			if (nodePair_thisLink.getFirst() == localRepairPoint) break;

			linkSpareCapacityMap.put(linkId, linkSpareCapacityMap.get(linkId) - x_p);
			if (nodePair_thisLink.getSecond() == localRepairPoint) break;
		}
		
		it = path.listIterator(path.size());
		while(it.hasPrevious())
		{
			long linkId = it.previous();
			Pair<Long, Long> nodePair_thisLink = currentNetPlan.getLinkOriginDestinationNodePair(linkId);
			if (nodePair_thisLink.getSecond() == mergePoint) break;

			linkSpareCapacityMap.put(linkId, linkSpareCapacityMap.get(linkId) - x_p);
			if (nodePair_thisLink.getFirst() == mergePoint) break;
		}
		
		return linkSpareCapacityMap;
	}

	/**
	 * Checks that a path is composed only of a sequence of links.
	 * 
	 * @param path Sequence of links
	 * @since 1.0
	 */
	private static void checkPath(List<Long> path)
	{
		for(long linkId : path)
			if (linkId < 0)
				throw new Net2PlanException("This algorithm does not allow routes to use protection segments");
	}
	
	/**
	 * Applies the restoration mechanism to all failing routes.
	 * 
	 * @param currentNetPlan Current network state
	 * @since 1.0
	 */
	private void rerouteFailingTraffic(NetPlan currentNetPlan)
	{
		/* Compute the link availability map (1 = link up, Double.MAX_VALUE = link down) */
		Map<Long, Double> linkAvailabilityMap = new LinkedHashMap<Long, Double>();
		for(long linkId : currentNetPlan.getLinkIds())
		{
			if (currentNetPlan.isLinkDown(linkId))
			{
				linkAvailabilityMap.put(linkId, Double.MAX_VALUE);
			}
			else
			{
				long originNodeId = currentNetPlan.getLinkOriginNode(linkId);
				long destinationNodeId = currentNetPlan.getLinkDestinationNode(linkId);

				if (currentNetPlan.isNodeDown(originNodeId) || currentNetPlan.isNodeDown(destinationNodeId))
					linkAvailabilityMap.put(linkId, Double.MAX_VALUE);
				else
					linkAvailabilityMap.put(linkId, 1.0);
			}
		}

		Map<Long, Pair<Long, Long>> linkMap = currentNetPlan.getLinkMap();

		/* Retrieve the set of routes that now are down */
		Set<Long> routesDown = new LinkedHashSet<Long>(currentNetPlan.getRoutesDown());
		for(long routeId : routesDown)
		{
			long ingressNodeId = currentNetPlan.getRouteIngressNode(routeId);
			long egressNodeId = currentNetPlan.getRouteEgressNode(routeId);
			if (currentNetPlan.isNodeDown(ingressNodeId) || currentNetPlan.isNodeDown(egressNodeId)) continue;

			/* First, try to use the original path */
			List<Long> seqLinks_primaryPath_thisRoute = currentNetPlan.getRouteOriginalSequenceOfLinks(routeId); checkPath(seqLinks_primaryPath_thisRoute);
			double x_p_thisRoute = currentNetPlan.getRouteOriginalCarriedTraffic(routeId);
			boolean primaryPathAvailable = currentNetPlan.isPathRestorable(seqLinks_primaryPath_thisRoute);
			if (primaryPathAvailable)
			{
				double x_p_newRoute = Math.min(x_p_thisRoute, TrafficComputationEngine.getPathBottleneckCapacity(currentNetPlan, seqLinks_primaryPath_thisRoute));
				if (x_p_newRoute >= x_p_thisRoute - PRECISION_FACTOR)
				{
					currentNetPlan.setRouteSequenceOfLinks(routeId, seqLinks_primaryPath_thisRoute);
					currentNetPlan.setRouteCarriedTraffic(routeId, x_p_newRoute);
					currentNetPlan.setRouteUp(routeId);
					continue;
				}
			}

			/* If fails, search for an alternate one according to the restoration mechanism */
			List<Long> seqLinks_currentPath_thisRoute = new LinkedList<Long>(currentNetPlan.getRouteSequenceOfLinks(routeId)); checkPath(seqLinks_currentPath_thisRoute);
			
			if (restorationMode.equals("linkRestoration"))
			{
				while(true)
				{
					long localRepairPoint = currentNetPlan.getFirstAvailableNodeBeforeFailures(seqLinks_currentPath_thisRoute);
					if (localRepairPoint == egressNodeId)
					{
						double x_p_newRoute = Math.min(x_p_thisRoute, TrafficComputationEngine.getPathBottleneckCapacity(currentNetPlan, seqLinks_currentPath_thisRoute));
						if (x_p_newRoute < x_p_thisRoute - PRECISION_FACTOR) break;

						currentNetPlan.setRouteSequenceOfLinks(routeId, seqLinks_currentPath_thisRoute);
						currentNetPlan.setRouteCarriedTraffic(routeId, x_p_newRoute);
						currentNetPlan.setRouteUp(routeId);
						break;
					}

					boolean pathFound = false;
					Set<Long> availableNodesFromNode = currentNetPlan.getAvailableNodesFromNode(localRepairPoint, seqLinks_currentPath_thisRoute);
					for(long mergePoint : availableNodesFromNode)
					{
						Map<Long, Double> linkSpareCapacityMap = actualLinkSpareCapacityMap(currentNetPlan, seqLinks_currentPath_thisRoute, localRepairPoint, mergePoint, x_p_thisRoute);
						List<Long> csp = GraphUtils.getCapacitatedShortestPath(linkMap, localRepairPoint, mergePoint, linkAvailabilityMap, linkSpareCapacityMap, x_p_thisRoute - PRECISION_FACTOR);
						if (csp.isEmpty()) continue;

						seqLinks_currentPath_thisRoute = currentNetPlan.getMergedPath(seqLinks_currentPath_thisRoute, csp);
						pathFound = true;
						break;
					}

					if (!pathFound) break;
				}
			}
			else
			{
				long localRepairPoint;
				long mergePoint;
				switch(restorationMode)
				{
					case "pathRestoration":
						localRepairPoint = ingressNodeId;
						mergePoint = egressNodeId;
						break;

					case "subpathRestoration":
						localRepairPoint = currentNetPlan.getFirstAvailableNodeBeforeFailures(seqLinks_currentPath_thisRoute);
						mergePoint = currentNetPlan.getFirstAvailableNodeAfterFailures(seqLinks_currentPath_thisRoute);
						break;

					case "fastReroute":
						localRepairPoint = currentNetPlan.getFirstAvailableNodeBeforeFailures(seqLinks_currentPath_thisRoute);
						mergePoint = egressNodeId;
						break;

					default:
						throw new Net2PlanException("Bad - Invalid restoration mode");
				}

				if (localRepairPoint == -1 || mergePoint == -1 || localRepairPoint == mergePoint) continue;

				List<Long> path;
				if (localRepairPoint == egressNodeId) /* A first available before a failure is egress node, then there is no failure in the current path, so use it */
				{
					if (mergePoint != ingressNodeId) throw new RuntimeException("Bad");

					path = seqLinks_currentPath_thisRoute;
				}
				else
				{
					Map<Long, Double> linkSpareCapacityMap = actualLinkSpareCapacityMap(currentNetPlan, seqLinks_currentPath_thisRoute, localRepairPoint, mergePoint, x_p_thisRoute);
					List<Long> csp = GraphUtils.getCapacitatedShortestPath(linkMap, localRepairPoint, mergePoint, linkAvailabilityMap, linkSpareCapacityMap, x_p_thisRoute - PRECISION_FACTOR);
					if (csp.isEmpty()) continue;

					path = currentNetPlan.getMergedPath(seqLinks_currentPath_thisRoute, csp);
				}

				double x_p_newRoute = Math.min(x_p_thisRoute, TrafficComputationEngine.getPathBottleneckCapacity(currentNetPlan, path));
				if (x_p_newRoute < x_p_thisRoute - PRECISION_FACTOR) continue;

				currentNetPlan.setRouteSequenceOfLinks(routeId, path);
				currentNetPlan.setRouteCarriedTraffic(routeId, x_p_newRoute);
				currentNetPlan.setRouteUp(routeId);
			}
		}
	}
	
	/**
	 * Tries to restore routes using an alternate path to their original path, if revertive mode 
	 * is enabled.
	 * 
	 * @param currentNetPlan Current network state
	 * @since 1.0
	 */
	private void restoreToPrimary(NetPlan currentNetPlan)
	{
		for(long routeId : currentNetPlan.getRouteIds())
		{
			boolean isRouteDown = currentNetPlan.isRouteDown(routeId);
			if (isRouteDown) continue;

			/* Check whether primary path is available */
			List<Long> seqLinks_primaryPath_thisRoute = currentNetPlan.getRouteOriginalSequenceOfLinks(routeId); checkPath(seqLinks_primaryPath_thisRoute);
			boolean primaryPathAvailable = currentNetPlan.isPathRestorable(seqLinks_primaryPath_thisRoute);
			if (!primaryPathAvailable) continue;

			/* Check whether route is using its primary path and carrying the original traffic, then skip it */
			List<Long> currentSeqLinks_thisRoute = new LinkedList<Long>(currentNetPlan.getRouteSequenceOfLinks(routeId));
			double x_p_primaryPath_thisRoute = currentNetPlan.getRouteOriginalCarriedTraffic(routeId);
			double current_x_p_thisRoute = currentNetPlan.getRouteCarriedTraffic(routeId);
			if (currentSeqLinks_thisRoute.equals(seqLinks_primaryPath_thisRoute) && x_p_primaryPath_thisRoute < current_x_p_thisRoute + PRECISION_FACTOR) continue;

			/* Try to restore route to its primary path */
			currentNetPlan.setRouteSequenceOfLinks(routeId, seqLinks_primaryPath_thisRoute);
			currentNetPlan.setRouteCarriedTraffic(routeId, x_p_primaryPath_thisRoute);

			/* If fails (some link is oversubscribed), leave it as was */
			if (currentNetPlan.getLinkMaximumUtilization() > 1 + PRECISION_FACTOR)
			{
				currentNetPlan.setRouteSequenceOfLinks(routeId, currentSeqLinks_thisRoute);
				currentNetPlan.setRouteCarriedTraffic(routeId, current_x_p_thisRoute);
			}
		}
	}
}
